!********************************************************************************
!*   Copyright(C) 2004-2012 Intel Corporation. All Rights Reserved.
!*   
!*   The source code, information  and  material ("Material") contained herein is
!*   owned  by Intel Corporation or its suppliers or licensors, and title to such
!*   Material remains  with Intel Corporation  or its suppliers or licensors. The
!*   Material  contains proprietary information  of  Intel or  its  suppliers and
!*   licensors. The  Material is protected by worldwide copyright laws and treaty
!*   provisions. No  part  of  the  Material  may  be  used,  copied, reproduced,
!*   modified, published, uploaded, posted, transmitted, distributed or disclosed
!*   in any way  without Intel's  prior  express written  permission. No  license
!*   under  any patent, copyright  or  other intellectual property rights  in the
!*   Material  is  granted  to  or  conferred  upon  you,  either  expressly,  by
!*   implication, inducement,  estoppel or  otherwise.  Any  license  under  such
!*   intellectual  property  rights must  be express  and  approved  by  Intel in
!*   writing.
!*   
!*   *Third Party trademarks are the property of their respective owners.
!*   
!*   Unless otherwise  agreed  by Intel  in writing, you may not remove  or alter
!*   this  notice or  any other notice embedded  in Materials by Intel or Intel's
!*   suppliers or licensors in any way.
!*
!********************************************************************************
!*   Content : MKL PARDISO Fortran-77 example
!*
!********************************************************************************
!C----------------------------------------------------------------------
!C Example program to show the use of the "PARDISO" routine
!C for symmetric linear systems
!C---------------------------------------------------------------------
!C This program can be downloaded from the following site:
!C www.pardiso-project.org
!C
!C (C) Olaf Schenk, Department of Computer Science,
!C University of Basel, Switzerland.
!C Email: olaf.schenk@unibas.ch
!C
!C---------------------------------------------------------------------
      INCLUDE 'mkl_pardiso.f90'
      PROGRAM pardiso_unsym
      USE mkl_pardiso
      IMPLICIT NONE
      integer, parameter :: dp = kind(1.0d0)
!C      include 'mkl_pardiso.f77'
!C.. Internal solver memory pointer for 64-bit architectures
!C.. INTEGER*8 pt(64)
!C.. Internal solver memory pointer for 32-bit architectures
!C.. INTEGER*4 pt(64)
!C.. This is OK in both cases

!! 64-bit 
!      INTEGER*8 pt(64)
!!C.. All other variables
!      INTEGER maxfct, mnum, mtype, phase, n, nrhs, error, msglvl
!      INTEGER iparm(64)
!      INTEGER ia(6)
!      INTEGER ja(13)
!      REAL*8 a(13)
!      REAL*8 b(5)
!      REAL*8 x(5)
!      INTEGER i, idum(1)
!      REAL*8  ddum(1)

! 64-bit 
      TYPE(MKL_PARDISO_HANDLE), ALLOCATABLE  :: pt(:)
!C.. All other variables
      INTEGER maxfct, mnum, mtype, phase, n, nrhs, error, msglvl, nnz
      INTEGER, allocatable :: iparm(:)
      INTEGER, allocatable :: ia(:)
      INTEGER, allocatable :: ja(:)
      REAL(kind = dp), allocatable :: a(:)
      REAL(kind = dp), allocatable :: b(:)
      REAL(kind = dp), allocatable :: x(:)
      INTEGER i, idum(1)
      REAL(kind = dp)  ddum(1)
      
      character(12) :: strnum
      integer :: iunit      
      integer :: ii, j, k
        
        !call mkl_set_dynamic(0)
        !call mkl_set_num_threads(12)
      
      do ii = 0, 3
          
        iunit = 21
      
        nrhs = 1 
        maxfct = 1 
        mnum = 1
      
!C   Allocate array
        !n = 5 
        !nnz = 13
        !if(.not. allocated(ia)) then
        !    ALLOCATE( ia ( n + 1 ) )    
        !end if
        !ia = (/1,4,6,9,12,14/)  
        
        write(strnum, *) ii
        open(unit = iunit, file = "ia_"//trim(adjustl(strnum))//".txt")
        read(iunit,*) n
        n = n - 1
        if(.not. allocated(ia)) then
            allocate( ia ( n + 1 ) )    
        end if
        if(size(ia,1) /= n + 1) then
            deallocate(ia)
            allocate(ia(n + 1))
        end if
        
        do i = 1, n + 1
            read(iunit,*) ia(i)
        end do
        close(iunit)
        write(*,*) "read ia finished, size of ia", size(ia,1)
        
        
        !if(.not. allocated(ja)) then
        !    ALLOCATE( ja ( nnz ) )
        !end if
        !ja = (/1,    2,          4,           &   !1
        !       1,    2,                       &   !2
        !                   3,    4,    5,     &   !3
        !       1,          3,    4,           &   !4
        !             2,                5/)        !5
        
        open(unit = iunit, file = "ja_"//trim(adjustl(strnum))//".txt")
        read(iunit,*) nnz
        if(.not. allocated(ja)) then
            allocate( ja ( nnz) )    
        end if
        if(size(ja,1) /= nnz) then
            deallocate(ja)
            allocate(ja(nnz))
        end if
        
        do i = 1, nnz
            read(iunit,*) ja(i)
        end do
        close(iunit)  
        write(*,*) "read ja finished. size of ja", size(ja,1)
        
        !if(.not. allocated(a)) then
        !    ALLOCATE( a ( nnz ) )
        !end if
        !a = (/1.d0,-1.d0,      -3.d0,           &   !1
        !     -2.d0, 5.d0,                       &   !2
        !                  4.d0, 6.d0, 4.d0,     &   !3
        !     -4.d0,       2.d0, 7.d0,           &   !4
        !            8.d0,            -5.d0/)        !5
        open(unit = iunit, file = "a_"//trim(adjustl(strnum))//".txt")
        read(iunit,*) nnz
        if(.not. allocated(a)) then
            allocate( a ( nnz) )    
        end if
        if(size(a,1) /= nnz) then
            deallocate(a)
            allocate(a(nnz))
        end if
        
        do i = 1, nnz
            read(iunit,*) a(i)
        end do
        close(iunit) 
        write(*,*) "read a finished, size of a", size(a,1)
        
        
        !if(.not. allocated(b)) then
        !    allocate(b(n))
        !end if
        open(unit = iunit, file = "b_"//trim(adjustl(strnum))//".txt")
        read(iunit,*) n
        if(.not. allocated(b)) then
            allocate(b (n))    
        end if
        if(size(b,1) /= n ) then
            deallocate(b)
            allocate(b(n))
        end if        
        do i = 1, n
            read(iunit,*) b(i)
        end do
        close(iunit)        
        write(*,*) "read b finished, size of b", size(b,1)
        
        write(*,*) "n:", n, ", nnz:", nnz
        
        !check data
        !do i = 1, n
        !    do j = ia(i), ia(i + 1)-1 
        !        do k = j + 1, ia(i + 1)-1
        !            if(ja(j)> ja(k)) then
        !                write(*,*) "error in ja value in row:", i
        !            end if
        !        end do
        !    end do
        !end do
        
        
        if(.not. allocated(x)) then
            allocate(x(n))
        end if
        if (size(x, 1) /= n) then
            deallocate(x)
            allocate(x(n))
        end if
        
        if(.not. allocated(iparm)) then
            allocate(iparm(64))
        end if
        
      
!C.. Fill all arrays containing matrix data.
!      DATA n /5/, nrhs /1/, maxfct /1/, mnum /1/
!      DATA ia /1,4,6,9,12,14/
      !DATA ja                             &
      ! /   1,    2,          4,           &   !1
      !     1,    2,                       &   !2
      !                 3,    4,    5,     &   !3
      !     1,          3,    4,           &   !4
      !           2,                5/         !5
      !DATA a                              &
      ! /1.d0,-1.d0,      -3.d0,           &   !1
      ! -2.d0, 5.d0,                       &   !2
      !              4.d0, 6.d0, 4.d0,     &   !3
      ! -4.d0,       2.d0, 7.d0,           &   !4
      !        8.d0,        0   -5.d0/         !5
!C..
!C.. Set up PARDISO control parameter
!C..
      do i = 1, 64
         iparm(i) = 0
      end do
      iparm(1) = 1 ! no solver default
      iparm(2) = 2 ! fill-in reordering from METIS
      iparm(3) = 0 ! numbers of processors      
      iparm(4) = 0 ! no iterative-direct algorithm
      iparm(5) = 0 ! no user fill-in reducing permutation
      iparm(6) = 0 ! =0 solution on the first n compoments of x
      iparm(7) = 0 ! not in use
      iparm(8) = 9 ! numbers of iterative refinement steps
      iparm(9) = 0 ! not in use
      iparm(10) = 13 ! perturbe the pivot elements with 1E-13
      iparm(11) = 1 ! use nonsymmetric permutation and scaling MPS
      iparm(12) = 0 ! not in use
      iparm(13) = 1 ! maximum weighted matching algorithm is switched-on (default for non-symmetric)
      iparm(14) = 0 ! Output: number of perturbed pivots
      iparm(15) = 0 ! not in use
      iparm(16) = 0 ! not in use
      iparm(17) = 0 ! not in use
      iparm(18) = -1 ! Output: number of nonzeros in the factor LU
      iparm(19) = -1 ! Output: Mflops for LU factorization
      iparm(20) = 0 ! Output: Numbers of CG Iterations
      iparm(27) = 1
      error = 0 ! initialize error flag
      msglvl = 1 ! print statistical information
      mtype = 11 ! real unsymmetric
!C.. Initiliaze the internal solver memory pointer. This is only
!C necessary for the FIRST call of the PARDISO solver.
      if(.not. allocated(pt)) then
        allocate(pt(64))
      end if
      do i = 1, 64
         !pt(i) = 0
         pt( i )%DUMMY =  0 
      end do
!C.. Reordering and Symbolic Factorization, This step also allocates
!C all memory that is necessary for the factorization
      phase = 11 ! only reordering and symbolic factorization
      CALL pardiso (pt, maxfct, mnum, mtype, phase, n, a, ia, ja,     &
        idum, nrhs, iparm, msglvl, ddum, ddum, error)
      WRITE(*,*) 'Reordering completed ... '
      IF (error .NE. 0) THEN
         WRITE(*,*) 'The following ERROR was detected: ', error
         STOP 1
      END IF
      WRITE(*,*) 'Number of nonzeros in factors = ',iparm(18)
      WRITE(*,*) 'Number of factorization MFLOPS = ',iparm(19)
!C.. Factorization.
      phase = 22 ! only factorization
      CALL pardiso (pt, maxfct, mnum, mtype, phase, n, a, ia, ja,     &
       idum, nrhs, iparm, msglvl, ddum, ddum, error)
      WRITE(*,*) 'Factorization completed ... '
      IF (error .NE. 0) THEN
         WRITE(*,*) 'The following ERROR was detected: ', error
         STOP 1
      ENDIF
!C.. Back substitution and iterative refinement
      iparm(8) = 2 ! max numbers of iterative refinement steps
      phase = 33 ! only factorization 
      !do i = 1, n
      !   b(i) = 1.d0
      !end do      
      CALL pardiso (pt, maxfct, mnum, mtype, phase, n, a, ia, ja,     &
       idum, nrhs, iparm, msglvl, b, x, error)
      WRITE(*,*) 'Solve completed ... '
      WRITE(*,*) 'The solution of the system is '
      !DO i = 1, n
      !   WRITE(*,*) ' x(',i,') = ', x(i)
      !END DO
      
      open(unit = iunit, file = "x_"//trim(adjustl(strnum))//".txt")
      write(iunit, *) n
      do i = 1, n
          write(iunit, *)  x(i)
      end do
      close(iunit)
      
!C.. Termination and release of memory
      phase = -1 ! release internal memory
      CALL pardiso (pt, maxfct, mnum, mtype, phase, n, ddum, idum,    &
       idum, idum, nrhs, iparm, msglvl, ddum, ddum, error)
      
      write(*,*) "============ solve ", ii, " finished =================="
      
      end do
      
        IF ( ALLOCATED( ia ) )      DEALLOCATE( ia )
        IF ( ALLOCATED( ja ) )      DEALLOCATE( ja )
        IF ( ALLOCATED( a ) )       DEALLOCATE( a )
        IF ( ALLOCATED( b ) )       DEALLOCATE( b )
        IF ( ALLOCATED( x ) )       DEALLOCATE( x )
        IF ( ALLOCATED( iparm ) )   DEALLOCATE( iparm )
      
      END

